// --8<--8<--8<--8<--
//
// Copyright (C) 2008 Smithsonian Astrophysical Observatory
//
// This file is part of fits2rdb
//
// fits2rdb is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or (at
// your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// -->8-->8-->8-->8--

#include <cmath>
#include <algorithm>
#include <pcre.h>
#include <suplib/str.h>
#include <tracefctxx/TraceFct.h>
using namespace std;
#include "Fits2Rdb.h"

Fits2Rdb::Fits2Rdb( clo::parser& parser ) {
  TraceFct tf( "Fits2Rdb::Fits2Rdb()" );

  const clo::options& options       = parser.get_options( );
  const vector<string>& non_options = parser.get_non_options( );

  string input, output, extname, arraysep;
  int extnum;
  try {
    input     = options.input;
    output    = options.output;
    extname   = options.extname;
    extnum    = options.extnum;
    arraysep  = options.arraysep;

    const char* delim = " \t\n,";
    for ( int idx = 0; idx < options.column.size( ); idx++ )
      suplib::tok( _column,   options.column[idx],   delim );
    for ( int idx = 0; idx < options.rcolumn.size( ); idx++ )
      suplib::tok( _rcolumn,  options.rcolumn[idx],  delim );
    for ( int idx = 0; idx < options.vcolumn.size( ); idx++ )
      suplib::tok( _vcolumn,  options.vcolumn[idx],  delim );
    for ( int idx = 0; idx < options.rvcolumn.size( ); idx++ )
      suplib::tok( _rvcolumn, options.rvcolumn[idx], delim );

    _comment   = options.comment;
    _trim      = options.trim;
    _boolvalue['T'] = options.truevalue;
    _boolvalue['F'] = options.falsevalue;
    _boolvalue['0'] = options.invalidvalue;
    _byterep   = options.byterep;
    _maxrow    = options.maxrow;
    _colformat = options.colformat;
    _initDefaultFormat( );
    for ( map<string,string>::const_iterator itr = options.defformat.begin( );
	  itr != options.defformat.end( ); 
	  itr++ )
      _defformat[itr->first] = itr->second;

    if ( non_options.size( ) ) {
      vector<string>::const_iterator file_itr = non_options.begin( );
      input = *file_itr;

      if ( ++file_itr != non_options.end( ) ) {
	output = *file_itr;

	if ( ++file_itr != non_options.end( ) )
	  tf.exit( EXIT_FAILURE, "More than 2 non-option arguments on the commandline '%s'\n", file_itr->c_str( ), NULL );
      }
    }
  } catch ( clo::option_error& oe ) {
    tf.exit( EXIT_FAILURE, oe.what( ), NULL );

  }

  int try_again = 0;
  char* filename = new char[ 
    input.size()
    + 
    extname.size()
    + 
    (int)(log((double)abs(extnum))+1)
    +
    5 // for the '[', ']' characters around EXTNAME and EXTNUM and '\0'...
  ];
  if ( extname.size() ) {
    if ( -1 != extnum ) {
      sprintf( filename, "%s[%s][%d]", input.c_str(), extname.c_str(), extnum );
    } else {
      sprintf( filename, "%s[%s]", input.c_str(), extname.c_str() );
    }

  } else if ( -1 != extnum ) {
    sprintf( filename, "%s[%d]", input.c_str(), extnum );

  } else {
    // user requested a file with no additional info about which
    // header... make special efforts to open it.
    try_again = 1;
    sprintf( filename, "%s", input.c_str() );

  }

  // Let's try to quiet fitsy down, shall we?
  setgerror( 0 );

  // user specified the file as well as an extension
  // assume they know what they are asking for and only try to open it
  // once...
  if ( ! try_again ) {
    char mode[] = "r";
    if ( ! (_fun = FunOpen( filename, mode, NULL )) ) { 
      
      tf.exit( EXIT_FAILURE, "Error opening '%s' for input\n", filename, NULL );
      
    }
  } else {
    // user only specified a file and not an extension.
    // try to open it and let FUNTOOLS determine which extension to use.
    char mode[] = "r";
    if ( !(_fun = FunOpen( filename, mode, NULL )) ) {
      // that didn't work, maybe there was no EXTNAME keyword...
      // so let's open up each extension til we find one that is a table.
      int ext_index = 0;
      int ftype = FUN_UNKNOWN;
      while ( FUN_TABLE != ftype ) {
	if ( _fun )
	  FunClose( _fun );

	sprintf( filename, "%s[%d]", input.c_str(), ext_index++ );
        char mode[] = "r";
	if ( !(_fun = FunOpen( filename, mode, NULL )) ) {
	  // this means either there is no such file or that the file
	  // has no FUN_TABLE extensions.
	  tf.message( "Error trying to find a default extension in '%s'\n", input.c_str( ), NULL );
	  tf.exit( EXIT_FAILURE, "Does this file exist?", input.c_str( ), NULL );
	}
	FunInfoGet( _fun, FUN_TYPE, &ftype, NULL );
      }
    }
    if ( !_fun ) { 
      tf.exit( EXIT_FAILURE, "Error opening '%s' for input\n", filename, NULL );
    }
  }

  delete [] filename;

  _ofileptr =
    ( 
      ( "stdout" == output ) 
      ?      
      ( _ofileptr = stdout )
      :
      ( 
	( "stderr" == output )
	?
	( _ofileptr = stderr )
	:
	( fopen( output.c_str(), "w" ) )
      )
    );

  if ( ! _ofileptr )
    tf.exit( EXIT_FAILURE, "Error opening '%s' for output\n", output.c_str( ), NULL );

  _arraysep = new char[arraysep.size()+1];
  strcpy( _arraysep, arraysep.c_str() );

}

Fits2Rdb::~Fits2Rdb( void ) {
  TraceFct tf( "Fits2Rdb::~Fits2Rdb()" );

  if ( _ofileptr )
    fclose( _ofileptr );

  if ( _arraysep )
       delete [] _arraysep;

  if ( _sep )
    delete [] _sep;

  if ( _print )
    delete [] _print;

  if ( _type )
    delete [] _type;

  if ( _mode )
    delete [] _mode;

  if ( _offset )
    delete [] _offset;

  if ( _n )
    delete [] _n;

  if ( _width )
    delete [] _width;

  if ( _fun )
    FunClose( _fun );

}

void Fits2Rdb::findColumns( void ) {
  TraceFct tf( "Fits2Rdb::findColumns()" );

  char* name;
  int type;
  int mode;
  int offset;
  int n;
  int width;

  try { 
    for ( vector<string>::iterator vitr = _column.begin(); vitr != _column.end(); vitr++ )
      for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
	*sitr = tolower( *sitr );
    for ( vector<string>::iterator vitr = _vcolumn.begin(); vitr != _vcolumn.end(); vitr++ )
      for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
	*sitr = tolower( *sitr );
  } catch ( clo::option_error& oe ) {
    tf.exit( EXIT_FAILURE, oe.what( ), NULL );
  }
  int addit = ( 0 == _column.size( ) && 0 == _rcolumn.size( ) );

  vector<pcre*> re;
  vector<pcre*> vre;
  const char* error = NULL;
  int erroffset = 0;

  for ( vector<string>::iterator itr = _rcolumn.begin( ); itr != _rcolumn.end( ); itr++ )
    re.push_back( pcre_compile( itr->c_str( ), PCRE_CASELESS, &error, &erroffset, NULL ) );
  
  for ( vector<string>::iterator itr = _rvcolumn.begin( ); itr != _rvcolumn.end( ); itr++ )
    vre.push_back( pcre_compile( itr->c_str( ), PCRE_CASELESS, &error, &erroffset, NULL ) );

  FunInfoGet( _fun, FUN_ROWSIZE, &_rowsize, FUN_NCOL, &_ncols, NULL );
  _ncols -= 2;

  _print  = new int[_ncols];
  _sep    = new char[_ncols];
  _type   = new int[_ncols];
  _mode   = new int[_ncols];
  _offset = new int[_ncols];
  _n      = new int[_ncols];
  _width  = new int[_ncols];

  for ( int idx = 0; idx < _ncols; idx++ ) {
    
    FunColumnLookup( _fun, NULL, idx, &name, &type, &mode, &offset, &n, &width );
    for ( int jdx = 0; name[jdx] != '\0'; jdx++ )
      name[jdx] = tolower( name[jdx] );

    if ( 'A' == type || 'B' == type ) {
      int got = 0;
      char TFORM[] = "TFORM";
      char* tform = FunParamGets( _fun, TFORM, idx+1, NULL, &got );
      char *ptr1 = NULL, *ptr2 = NULL;
      long d1 = strtol( tform, &ptr1, 10 );
      long d2 = strtol( ++ptr1, &ptr2, 10 );
      if ( ptr1 != ptr2 )
	n = (int) ( d1 / d2 );
      else 
	n = 1;
      if ( tform )
	free( tform );
    }

    if ( addit ) {
      _print[idx] = 1;
    } else {
      _print[idx] = 0;

      if ( _column.end( ) != find( _column.begin( ), 
				  _column.end( ),
				  string( name ) ) ) {
	_print[idx] = 1;
      }

      for ( vector<pcre*>::iterator itr = re.begin(); itr != re.end(); itr++ ) 
	if ( 0 <= pcre_exec( *itr, NULL, name, 8, 0, 0, NULL, 0 ) ) {
	  _print[idx] = 1;
	  itr = re.end()-1;
	}
    }
     
    if ( _vcolumn.end( ) != find( _vcolumn.begin( ), 
				 _vcolumn.end( ),
				 string( name ) ) ) {
      _print[idx] = 0;
    }

    for ( vector<pcre*>::iterator itr = vre.begin(); itr != vre.end(); itr++ ) {
      if ( 0 <= pcre_exec( *itr, NULL, name, 8, 0, 0, NULL, 0 ) ) {
	_print[idx] = 0;
	itr = vre.end()-1;
      }
    }

    _sep[idx]    = '\t';
    _name[idx]   = name;
    _type[idx]   = type;
    _mode[idx]   = mode;
    _offset[idx] = offset;
    _n[idx]      = n;
    _width[idx]  = width;
  }
    
  for ( vector<pcre*>::iterator itr = re.begin( ); itr != re.end( ); itr++ )
    free( *itr );
    
  for ( vector<pcre*>::iterator itr = vre.begin( ); itr != vre.end( ); itr++ )
    free( *itr );

  for ( int idx = _ncols-1; idx >= 0; idx-- ) {
    if ( _print[idx] ) {

      _sep[idx] = '\n';
      idx = -1;

    }
  }
}

void Fits2Rdb::findColumns2( void ) {
  TraceFct tf( "Fits2Rdb::findColumns2()" );

  char* name;
  int type;
  int mode;
  int offset;
  int n;
  int width;

  for ( vector<string>::iterator vitr = _column.begin(); vitr != _column.end(); vitr++ )
    for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
      *sitr = tolower( *sitr );
  for ( vector<string>::iterator vitr = _rcolumn.begin(); vitr != _rcolumn.end(); vitr++ )
    for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
      *sitr = tolower( *sitr );
  for ( vector<string>::iterator vitr = _vcolumn.begin(); vitr != _vcolumn.end(); vitr++ )
    for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
      *sitr = tolower( *sitr );
  for ( vector<string>::iterator vitr = _rvcolumn.begin(); vitr != _rvcolumn.end(); vitr++ )
    for ( string::iterator sitr = vitr->begin(); sitr != vitr->end(); sitr++ )
      *sitr = tolower( *sitr );
  
  FunInfoGet( _fun, FUN_ROWSIZE, &_rowsize, FUN_NCOL, &_ncols, NULL );
  _ncols -= 2;

  _print  = new int[_ncols];
  _sep    = new char[_ncols];
  _type   = new int[_ncols];
  _mode   = new int[_ncols];
  _offset = new int[_ncols];
  _n      = new int[_ncols];
  _width  = new int[_ncols];

  for ( int idx = 0; idx < _ncols; idx++ ) {
    
    FunColumnLookup( _fun, NULL, idx, &name, &type, &mode, &offset, &n, &width );
    for ( int jdx = 0; name[jdx] != '\0'; jdx++ )
      name[jdx] = tolower( name[jdx] );

    if ( 'A' == type || 'B' == type ) {
      int got = 0;
      char TFORM[] = "TFORM";
      char* tform = FunParamGets( _fun, TFORM, idx+1, NULL, &got );
      char *ptr1 = NULL, *ptr2 = NULL;
      long d1 = strtol( tform, &ptr1, 10 );
      long d2 = strtol( ++ptr1, &ptr2, 10 );
      if ( ptr1 != ptr2 ) {
	n = (int) ( d1 / d2 );
	n = n ? n : 1;
      } else 
	n = 1;
      if ( tform )
	free( tform );
    }

    _print[idx]  = 0;
    _sep[idx]    = '\t';
    _name.insert( _name.end( ), name );
    _type[idx]   = type;
    _format.insert( _format.end( ), _defformat[string(1,_type[idx])] );
    if ( _colformat.end( ) != _colformat.find( _name[idx] ) )
      _format[idx] = _colformat[_name[idx]];
    _mode[idx]   = mode;
    _offset[idx] = offset;
    _n[idx]      = n;
    _width[idx]  = width;
  }

  suplib::colselect( _name, _column, _rcolumn, _vcolumn, _rvcolumn, _ocolumn );

  bool last = true;
  for ( int idx = _ncols-1; idx >= 0; idx-- ) { 
    if ( _ocolumn.end( ) != find( _ocolumn.begin( ), _ocolumn.end( ), _name[idx] ) ) {
      _print[idx] = 1;
      if ( last ) {
	_sep[idx] = '\n';
	last = false;

      }
    }
  }
}

void Fits2Rdb::convertHeader( void ) {
  TraceFct tf( "Fits2Rdb::convertHeader()" );

  FITSHead fhead;
  char name[FT_WORDLEN+1], value[FT_CARDLEN+1], comm[FT_CARDLEN+2+1];
  FITSType type;
  int index, ncomment = 0, nhistory = 0, longstrn = 0;

  char LONGSTRN[] = "LONGSTRN";
  FunParamGets( _fun, LONGSTRN, 0, NULL, &longstrn );
  FunInfoGet( _fun, FUN_HEADER, &fhead, NULL );
  for ( int idx = 0; idx < fhead->ncard-1; idx++ ) {
    memset( name, '\0', FT_WORDLEN+1 );
    memset( value, '\0', FT_CARDLEN+1 );
    memset( comm, '\0', FT_CARDLEN+2+1 );

    strncpy( name, fhead->cards[idx].c, FT_WORDLEN );
    ft_cardpar( &(fhead->cards[idx]), &type, value, &index, comm );
    if ( _trim ) {
      str_trim( value );
      str_trunc( value );
      str_trim( comm );
      str_trunc( comm );
    }

    if ( FT_COMMENT == type ) {
      if ( _trim ) {
	str_trim( name );
	str_trunc( name );
      }

      fprintf( _ofileptr, "#:%s_%d=%s\n", 
	       name,
	       ((0==strcmp("COMMENT",name))?(ncomment++):(nhistory++)), 
	       value
	       );

    } else {
      int len = 0;
      if ( longstrn ) {
	len = strlen( value );
      }
	
      if ( longstrn && len && '&' == value[len-1] ) {
	value[len-1] = '\0';
	fprintf( _ofileptr, "#:%s=%s%s%s", 
		 name, 
		 value, 
		 (comm[0])?(" /"):("\0"),
		 (comm[0])?(comm):("\0")
		 );
	while ( idx+1 < fhead->ncard-1 && 0 == strncmp( "CONTINUE", fhead->cards[idx+1].c, FT_WORDLEN ) ) { 
	  idx++;
	  ft_cardpar( &(fhead->cards[idx]), &type, value, &index, comm );
	  if ( _trim ) {
	    str_trim( value );
	    str_trunc( value );
	    str_trim( comm );
	    str_trunc( comm );
	  }
	  unquote( value );
	  fprintf( _ofileptr, "%s%s%s", 
		   value,
		   (comm[0])?(" /"):("\0"),
		   (comm[0])?(comm):("\0")
		   );
	}
	fprintf( _ofileptr, "\n" );

      } else {
	fprintf( _ofileptr, "#:%s=%s%s%s\n", 
		 name, 
		 value, 
		 (comm[0])?(" /"):("\0"),
		 (comm[0])?(comm):("\0")
		 );
      }
    }
  }
}

void Fits2Rdb::convertTable( void ) {
  TraceFct tf( "Fits2Rdb::convertTable()" );

  try { 
    if ( _comment )
      convertHeader( );

  } catch ( clo::option_error& oe ) {
    tf.exit( EXIT_FAILURE, oe.what( ), NULL );

  } catch ( exception& e ) {
    tf.exit( EXIT_FAILURE, e.what( ), NULL );

  }

  char* funstruct;
  char* row = NULL;
  char* ptr;
  char** fld = new char*[_ncols];
  int* sz = new int[_ncols];
  int nrows;
  int noutputcols = 0;
  char type;

  if ( ! (funstruct = (char *)calloc( _rowsize*_maxrow, sizeof(char) )) ) {
    while ( _rowsize > 2 && ! funstruct ) { 
      _rowsize /= 2;
      tf.message( "Error allocating row buffer, reducing maxrow to %d\n", _maxrow );
      funstruct = (char *)calloc( _rowsize*_maxrow, sizeof(char) );
    }
  }

  if ( ! funstruct ) {
    tf.exit( EXIT_FAILURE, "Error allocating row buffer, no space\n" );
  }

  for ( int idx = 0; idx < _ncols; idx++ ) {
    if ( _print[idx] ) {
      fprintf( _ofileptr, "%s%c", _name[idx].c_str(), _sep[idx] );
      noutputcols++;
    }
  }
  for ( int idx = 0; idx < _ncols; idx++ ) {
    if ( _print[idx] ) {
      if ( 'A' == _type[idx] || 
	   'B' == _type[idx] || 
	   ( 'L' == _type[idx] && 
	     ( 
	       _boolvalue['T'].npos !=
	       _boolvalue['T'].find_first_not_of( "1234567890" ) 
	       ||
	       _boolvalue['F'].npos != 
	       _boolvalue['F'].find_first_not_of( "1234567890" ) 
	       ||
	       _boolvalue['0'].npos != 
	       _boolvalue['0'].find_first_not_of( "1234567890" ) 
	     ) 
	   ) 
	 ) {
	type = 'S';
	sz[idx]  = _width[idx] / _n[idx];
	fld[idx] = new char[sz[idx]+1];
	fld[idx][sz[idx]] = '\0';

      } else {
	type = 'N';

      }
      
      int got = 0;
      char TUNIT[] = "TUNIT";
      char* units = FunParamGets( _fun, TUNIT, idx+1, NULL, &got );
      if ( got ) {
	fprintf( _ofileptr, "%c [%s]%c",
		 type, units, _sep[idx] );
      } else {
	fprintf( _ofileptr, "%c%c",
		 type, _sep[idx] );
      }
      if ( units )
	free( units );
    }
  }

  if ( noutputcols ) {
    while( FunTableRowGet( _fun, funstruct, _maxrow, NULL, &nrows ) ) {
      row = funstruct;
      for ( int jdx = 0; jdx < nrows; jdx++ ) { 
	for ( int idx = 0; idx < _ncols; idx++ ) {
	  if ( _print[idx] ) {
	    ptr = row + _offset[idx];
	  
	    // From the FITS standard, 
	    //  Definition of the Flexible Image Transport System (FITS) 
	    //  March 29, 1999
	    //  NOST 100-2.0
	    //  NASA/Science Office of Standards and Technology 
	    //  Code 633.2 
	    //  NASA Goddard Space Flight Center 
	    //  Greenbelt MD 20771 
	    //  USA 
	    // expect these datatypes....
	    
	    switch ( _type[idx] ) {
	    case 'A':
	      strncpy( fld[idx], (char*)ptr, sz[idx] );
	      for ( int j = 0; j < _width[idx]-sz[idx]; j += sz[idx] ) {
		if ( _trim ) {
		  str_trim( fld[idx] );
		  str_trunc( fld[idx] );
		}
		
		fprintf( _ofileptr, _format[idx].c_str( ), fld[idx] );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sz[idx];
		strncpy( fld[idx], (char*)ptr, sz[idx] );
	      }
	      if ( _trim ) {
		str_trim( fld[idx] );
		str_trunc( fld[idx] );
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), fld[idx] );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'B':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		if ( ! isspace( *(unsigned char*)ptr ) || ! _trim )
		  fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned char*)ptr) );
		ptr += sizeof(unsigned char);
	      }
	      if ( ! isspace( *(unsigned char*)ptr ) || ! _trim )
		fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned char*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'L':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), _boolvalue[*((unsigned char*)ptr)].c_str() );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(unsigned char);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), _boolvalue[*((unsigned char*)ptr)].c_str() );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'I':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((short*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(short);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((short*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'U':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned short*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(unsigned short);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned short*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'J':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((int*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(int);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((int*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'V':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned int*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(unsigned int);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((unsigned int*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'E':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((float*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(float);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((float*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'D':
	      for ( int j = 0; j < _n[idx]-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), *((double*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr += sizeof(double);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), *((double*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'X':
	      for ( int j = 0; j < ((_n[idx]+7)/8)-1; j++ ) {
		fprintf( _ofileptr, _format[idx].c_str( ), (unsigned int)*((unsigned char*)ptr) );
		fprintf( _ofileptr, "%s", _arraysep );
		ptr+=sizeof(unsigned char);
	      }
	      fprintf( _ofileptr, _format[idx].c_str( ), (unsigned int)*((unsigned char*)ptr) );
	      putc( _sep[idx], _ofileptr );
	      break;
	    case 'P':
	      tf.exit( EXIT_FAILURE, "The Array Descriptor datatype, '%c' is not support by this program currently...\n", (char)_type[idx], NULL );
	    default:
	      tf.exit( EXIT_FAILURE, "Unexpected datatype '%c'\n", (char)_type[idx], NULL );
	    }
	  }
	}
	row += _rowsize;
      }
      
      //      free( funstruct );
      //      funstruct = NULL;
      
    }
  }

  free( funstruct );
  delete [] sz;
  for ( int idx = 0; idx < _ncols; idx++ )
    if ( 'A' == _type[idx] || 
	 'B' == _type[idx] || 
	 ( 'L' == _type[idx] && 
	   ( 
	     _boolvalue['T'].npos !=
	     _boolvalue['T'].find_first_not_of("1234567890" ) 
	     ||
	     _boolvalue['F'].npos != 
	     _boolvalue['F'].find_first_not_of( "1234567890" ) 
	     ||
	     _boolvalue['0'].npos != 
	     _boolvalue['0'].find_first_not_of( "1234567890" ) 
	   ) 
	 )
	) 
      delete [] fld[idx];
  delete [] fld;

}

void Fits2Rdb::_initDefaultFormat( void ) {

  // Get space to construct E and D from limits.h...
  char* fmt = new char[4 + int(log((double)DBL_DIG)+1)];

  _defformat[string(1,'L')] = "%s";
  _defformat[string(1,'X')] = "%x";
  _defformat[string(1,'B')] = "%c";
  _defformat[string(1,'I')] = "%d";
  _defformat[string(1,'U')] = "%d"; // Funtools type, not FITS Standard.
  _defformat[string(1,'J')] = "%d";
  _defformat[string(1,'V')] = "%d"; // Funtools type, not FITS Standard.
  _defformat[string(1,'A')] = "%s";
  sprintf( fmt, "%c%c%d%c", '%', '.', FLT_DIG, 'g' );
  _defformat[string(1,'E')] = fmt;
  sprintf( fmt, "%c%c%d%c", '%', '.', DBL_DIG, 'g' );
  _defformat[string(1,'D')] = fmt;
  _defformat[string(1,'C')] = "%s"; // Unsupported currently.
  _defformat[string(1,'M')] = "%s"; // Unsupported currently.
  _defformat[string(1,'P')] = "%s"; // Unsupported currently.

  delete [] fmt;

}
